''' Clase para facilitar la creación de demonios '''

import os
import sys
import log
import time
import signal
import daemon
import argparse
from enum import Enum
from pid import PidFile
from pathlib import Path
from typing import Callable

class Resultado(Enum):
	EXITO = 0
	ERROR = 1

	@classmethod
	def negar(cls, resultado: 'Resultado'):
		return cls.EXITO if resultado == cls.ERROR else cls.ERROR

class Demonio:
	def __init__(self, inicio: Callable, ciclo: Callable, archivo_pid: str, archivo_log: str):
		self.ruta = Path(__file__).resolve().parent
		self.archivo_pid = archivo_pid
		self.archivo_log = archivo_log

		# La funcion de inicio se ejecuta una sola vez al iniciar el proceso
		self.inicio = inicio

		# La funcion de ciclo se ejecuta en un ciclo infinito
		self.ciclo = ciclo

	@property
	def corriendo(self):
		return (self.ruta / self.archivo_pid).exists()

	def estatus(self) -> Resultado:
		if self.corriendo:
			print('El proceso se esta ejecutando')
			return Resultado.EXITO
		else:
			print('El proceso no se esta ejecutando')
			return Resultado.ERROR

	def iniciar(self) -> Resultado:
		if self.corriendo:
			return Resultado.negar(self.estatus())

		demonio = daemon.DaemonContext \
		(
			stdout= open(self.ruta / self.archivo_log, 'a+'),
			stderr= open(self.ruta / self.archivo_log, 'a+'),
			pidfile= PidFile(self.archivo_pid, self.ruta)
		)

		with demonio:

			# Se redirigen los logs
			log.configurar_logs()
			sys.stdout = log.LogHandle(log.INFO)
			sys.stderr = log.LogHandle(log.ERROR)

			# Se muestra el mensaje de inicio
			print('Proceso ejecutandose en segundo plano')
			print(f'ID del proceso: {os.getpid()}')

			# Se ejecuta la función de inicio
			self.inicio()

			# Se ejecuta la función de ciclo
			while True:
				try:
					self.ciclo()
				except KeyboardInterrupt:
					print('Proceso detenido por el usuario')
					return Resultado.EXITO
				except Exception as e:
					log.error('Proceso detenido debido a un error')
					log.error(e)
					return Resultado.ERROR

	def detener(self):
		if not self.corriendo:
			return self.estatus()

		# Se obtiene el PID del proceso
		with (self.ruta / self.archivo_pid).open() as archivo:
			pid = archivo.read()

		print('Deteniendo proceso...')

		# Se detiene el proceso
		while self.corriendo:
			try:
				os.kill(int(pid), signal.SIGINT)
				time.sleep(1)
			except:
				log.error('No se pudo detener el proceso')
				return Resultado.ERROR

		print('Proceso detenido')
		return Resultado.EXITO

	def interfaz(self):
		parser = argparse.ArgumentParser()
		parser.add_argument('accion', choices= ['estatus', 'iniciar', 'detener'])
		args = parser.parse_args()

		if args.accion == 'estatus':
			resultado = self.estatus()
		elif args.accion == 'iniciar':
			resultado = self.iniciar()
		elif args.accion == 'detener':
			resultado = self.detener()

		sys.exit(resultado.value)