import pandas as pd
import sqlalchemy as sa
from typing import List

# Módulos locales
from modelos import Conexion, RegistroKafka, GRUPO_EDAD, TIPO_VALOR
from modelos import Estado, Municipio, GrupoEdad, TipoValor, DetalleValor, ValorReal

def generar_dataframe(registros: List[RegistroKafka]):
	''' Genera un dataframe a partir de los valores de una lista de registros '''

	# Si no hay registros, se regresa un dataframe vacío
	if len(registros) == 0:
		return pd.DataFrame()

	# Se crea el dataframe con los registros y las columnas especificadas
	df = pd.DataFrame \
	(
		[registro.__dict__ for registro in registros],
		columns =
		[
			'ID_REGISTRO',
			'FECHA_ACTUALIZACION',
			'ENTIDAD_RES', 'MUNICIPIO_RES',
			'FECHA_SINTOMAS', 'FECHA_DEF',
			'EDAD', 'CLASIFICACION_FINAL'
		]
	)

	# Se convierten los campos de fecha
	df['FECHA_ACTUALIZACION'] = pd.to_datetime(df['FECHA_ACTUALIZACION'])
	df['FECHA_SINTOMAS'] = pd.to_datetime(df['FECHA_SINTOMAS'])
	df['FECHA_DEF'] = pd.to_datetime(df['FECHA_DEF'], format='AAAA-MM-DD', errors = 'coerce')

	# Función 'resultado' que interpreta la columna de 'CLASIFICACION_FINAL' y 'FECHA_DEF'
	def resultado(valor):
		if valor in [1, 2, 3]:
			return 'CONFIRMADO'
		elif valor in [4, 5]:
			return 'INVIABLE'
		elif valor in [6]:
			return 'SOSPECHOSO'
		elif valor in [7]:
			return 'NEGATIVO'
		else:
			return 'DESCONOCIDO'

	# Se crea una nueva columna 'RESULTADO' y se elimina la columna 'CLASIFICACION_FINAL'
	df['RESULTADO'] = df['CLASIFICACION_FINAL'].apply(resultado)
	del df['CLASIFICACION_FINAL']

	# Se crea una nueva columna 'GRUPO_EDAD' con los grupos de edad
	grupos = [17, 29, 39, 49, 59, 1000000]
	etiquetas = [str(grupo) for grupo in GRUPO_EDAD]
	df['GRUPO_EDAD'] = pd.cut(df['EDAD'], grupos, labels = etiquetas)

	# Se reemplazan todos los valores de 'NaT' por nulos
	df.replace({pd.NaT: None}, inplace = True)

	return df

def calcular_confirmados(registros: pd.DataFrame, tipo: TIPO_VALOR, fecha: pd.Timestamp):
	''' Calcula el número de confirmados en los últimos 30 días a partir de la fecha dada '''

	# Se filtran los datos positivos para la fecha dada según el pivote
	confirmados = pd.DataFrame()
	fecha_inicial = fecha - pd.Timedelta(days = 30)
	datos_periodo = registros[registros[tipo.value] >= fecha_inicial]
	positivos = datos_periodo[datos_periodo['RESULTADO'] == 'POSITIVO']

	# Se calcula el número de registros confirmados agrupados por entidad y municipio
	for grupo in GRUPO_EDAD:
		grupo_confirmados = positivos[positivos['GRUPO_EDAD'] == str(grupo)]
		grupo_confirmados = grupo_confirmados.groupby(['ENTIDAD_RES', 'MUNICIPIO_RES'])
		grupo_confirmados = grupo_confirmados.size().reset_index(name = 'CONFIRMADOS')
		grupo_confirmados['GRUPO_EDAD'] = str(grupo)
		confirmados = pd.concat([confirmados, grupo_confirmados])

	return confirmados

def obtener_foraneas_general(conexion: Conexion, datos: pd.DataFrame):
	'''
		Obtiene las llaves foráneas de los siguientes campos:
			- ENTIDAD_RES   -> 'fk_estado'
			- MUNICIPIO_RES -> 'fk_municipio'
			- GRUPO_EDAD    -> 'fk_grupo_edad'
			- TIPO_VALOR    -> 'fk_tipo_valor'
	'''

	# Se obtienen las llaves foráneas de los estados
	estados = pd.read_sql(conexion.sesion.query(Estado).statement, conexion.enlace)
	estados = estados[['id', 'id_dge']]
	estados.columns = ['fk_estado', 'ENTIDAD_RES']
	datos = pd.merge(datos, estados, on = 'ENTIDAD_RES', how = 'left')

	# Se obtienen las llaves foráneas de los municipios
	municipios = pd.read_sql(conexion.sesion.query(Municipio).statement, conexion.enlace)
	municipios = municipios[['id', 'id_dge', 'fk_estado']]
	municipios.columns = ['fk_municipio', 'MUNICIPIO_RES', 'fk_estado']
	datos = pd.merge(datos, municipios, on = ['fk_estado', 'MUNICIPIO_RES'], how = 'left')

	# Se obtienen las llaves foráneas de los grupos de edad
	grupos_edad = pd.read_sql(conexion.sesion.query(GrupoEdad).statement, conexion.enlace)
	grupos_edad = grupos_edad[['id', 'nombre']]
	grupos_edad.columns = ['fk_grupo_edad', 'GRUPO_EDAD']
	datos = pd.merge(datos, grupos_edad, on = 'GRUPO_EDAD', how = 'left')

	# Se obtienen las llaves foráneas de los tipos de valor
	tipos_valor = pd.read_sql(conexion.sesion.query(TipoValor).statement, conexion.enlace)
	tipos_valor = tipos_valor[['id', 'tipo']]
	tipos_valor.columns = ['fk_tipo_valor', 'TIPO_VALOR']
	datos = pd.merge(datos, tipos_valor, on = 'TIPO_VALOR', how = 'left')

	# Se eliminan los renglones cuyo municipio no se encuentran en la base de datos
	datos.dropna(subset = ['fk_municipio'], inplace = True)

	# Se asigna la columna de 'valor' con los valores de la columna 'CONFIRMADOS'
	datos['valor'] = datos['CONFIRMADOS']

	return datos

def obtener_foraneas_detalles(conexion: Conexion, datos: pd.DataFrame):
	''' Obtiene la llave foránea de la tabla de detalles '''

	# Se obtienen las llaves foráneas de la tabla de detalles
	detalle_valores = pd.read_sql(conexion.sesion.query(DetalleValor).statement, conexion.enlace)
	detalle_valores.rename(columns = {'id': 'fk_detalle_valor'}, inplace = True)
	detalle_valores['fecha'] = pd.to_datetime(detalle_valores['fecha'])
	interseccion = ['fecha', 'fk_municipio', 'fk_grupo_edad', 'fk_tipo_valor']
	return pd.merge(datos, detalle_valores, on = interseccion, how = 'left')

def cargar_valores_reales(conexion: Conexion, registros: pd.DataFrame, fecha: pd.Timestamp):
	''' Carga los valores reales en la base de datos '''

	datos = []

	# Se obtienen los casos confirmados para cada tipo de valor
	for tipo in TIPO_VALOR:
		confirmados = calcular_confirmados(registros, tipo, fecha)
		confirmados['TIPO_VALOR'] = str(tipo)
		datos.append(confirmados)

	# Se concatenan los datos de los diferentes tipos de valor y se agrega la fecha
	confirmados = pd.concat(datos)
	confirmados['fecha'] = fecha

	# Se obtienen las llaves foráneas generales
	confirmados = obtener_foraneas_general(conexion, confirmados)

	# Se insertan los datos de la tabla de detalles en la base de datos
	conexion.sesion.execute(sa.insert(DetalleValor), confirmados.to_dict(orient = 'records'))

	# Se obtienen las llaves foráneas de la tabla de detalles
	confirmados = obtener_foraneas_detalles(conexion, confirmados)

	# Se insertan los valores reales en la base de datos
	conexion.sesion.execute(sa.insert(ValorReal), confirmados.to_dict(orient = 'records'))