''' Contiene la clase 'Conexión' que representa la conexión de un consumidor de Kafka a un tópico '''

from typing import Optional, Callable, Any

# Confluent Kafka
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer
from confluent_kafka import Consumer, TopicPartition, OFFSET_STORED
from confluent_kafka.serialization import SerializationContext, MessageField

class Conexion:
	''' Representa la conexión de un consumidor de Kafka a un tópico '''

	def __init__ \
	(
		self,
		bootrap_servers: str,
		group_id: str,
		schema_registry_url: str,
		topico_kafka: str,
		particion_kafka: int,
		from_dict: Callable[[dict, Any], Optional[Any]]
	):

		# Se inicializa el esquema y el deserializador de Avro
		self.schema_registry_client = SchemaRegistryClient({'url': schema_registry_url})
		self.avro_deserializer = AvroDeserializer(self.schema_registry_client, from_dict = from_dict)

		# Se crea el consumidor de Kafka
		self.consumer_settings = \
		{
			'bootstrap.servers': bootrap_servers,
			'group.id': group_id,
			'auto.offset.reset': 'earliest'
		}
		self.consumer = Consumer(self.consumer_settings)

		# Se asigna el consumidor a la partición del tópico y se crea el contexto de serialización
		self.topic_partition = TopicPartition(topico_kafka, particion_kafka, OFFSET_STORED)
		self.serialization_context = SerializationContext(self.topic_partition.topic, MessageField.VALUE)
		self.consumer.assign([self.topic_partition])

		# Se usa para almacenar el indice local del primer mensaje leído por el consumidor de Kafka
		self.indice_local = None

	def cerrar(self):
		''' Cierra el consumidor de Kafka '''

		self.consumer.close()

	def reasignar(self):
		''' Reasigna el consumidor de Kafka si se ha perdido la conexión '''

		try:
			# Esta función devuelve un 'RuntimeError' si el consumidor de Kafka se ha desconectado
			self.consumer.consumer_group_metadata()

		except RuntimeError as e:
			# Basta con crear un nuevo consumidor de Kafka para reasignarlo
			self.consumer = Consumer(self.consumer_settings)
			self.consumer.assign([self.topic_partition])

	def mover_indice_remoto(self, indice: Optional[int]):
		''' Posiciona el consumidor de Kafka en el indice indicado '''

		# No se hace nada si el indice es nulo
		if indice is None:
			return

		self.reasignar()
		self.consumer.unassign()
		self.topic_partition.offset = indice
		self.consumer.commit(offsets=[self.topic_partition])
		self.consumer.assign([self.topic_partition])

	def indice_mensaje_remoto(self) -> Optional[int]:
		'''
			Obtiene el indice del ultimo mensaje leído por consumidor de Kafka
			- Si no se ha leído ningún mensaje, devuelve nulo
		'''

		self.reasignar()
		indice = self.consumer.position([self.topic_partition])[0].offset
		return indice if indice > 0 else None

	def regresar_indice_remoto(self):
		'''
			Regresa el indice remoto del consumidor para que coincida con el indice local
			- Se usa para revertir el flujo de datos al punto de partida (indice local)
		'''

		self.mover_indice_remoto(self.indice_local)

	def avanzar_indice_local(self):
		'''
			Avanza el indice local del consumidor para que coincida con el indice remoto
			- Se usa para actualizar el punto de partida (indice local) del flujo de datos
		'''

		self.indice_local = self.indice_mensaje_remoto()

	def obtener_registro(self) -> Optional[Any]:
		''' Obtiene el siguiente registro desde el consumidor de Kafka '''

		self.reasignar()

		# Se lee el siguiente mensaje del consumidor de Kafka
		mensaje = self.consumer.poll(1)

		# Se regresa nulo si no se ha leído ningún mensaje o si el mensaje contiene un error
		if mensaje is None or mensaje.error() is not None:
			return None

		# Se actualiza el indice del primer mensaje leído
		if self.indice_local is None:
			self.indice_local = self.indice_mensaje_remoto() - 1

		# Se deserializa el mensaje y se regresa el registro
		return self.avro_deserializer(mensaje.value(), self.serialization_context)