''' Contiene las clases de conexión y los modelos asociados a las bases de datos '''

import pandas as pd
from enum import Enum
from pathlib import Path
from typing import Optional

# SQLAlchemy
import sqlalchemy as sa
import sqlalchemy.orm as orm
from sqlalchemy.pool import NullPool

# Ruta del archivo actual
RUTA = Path(__file__).resolve().parent

# Clases base de SQLAlchemy
SQLite = orm.declarative_base()
PostgreSQL = orm.declarative_base()

# Diccionario de datos para estados y municipios
HOJA_ENTIDADES = 'Catálogo de ENTIDADES'
HOJA_MUNICIPIOS = 'Catálogo MUNICIPIOS'
ARCHIVO_DICCIONARIO = RUTA / 'datos/201128 Catalogos.xlsx'

# ---------------------- Conexiones a las bases de datos ----------------------

class Conexion:
	''' Representa una conexión genérica a una base de datos '''

	def __init__(self, cadena_conexion: str, metadata: sa.MetaData):
		# Se conecta a la base de datos y se crea la sesión de SQLAlchemy
		self.motor_sql = sa.create_engine(cadena_conexion, echo = False, poolclass = NullPool)
		self.sesion = orm.Session(bind = self.motor_sql, autoflush = False)

		# Inicializa la base de datos correspondiente con los modelos de SQLAlchemy
		metadata.create_all(self.motor_sql)

	@property
	def enlace(self):
		''' Regresa el enlace de la conexión de la transacción actual '''

		return self.sesion.connection()

	def cerrar(self):
		''' Cierra la sesión de SQLAlchemy '''

		self.sesion.close()

class ConexionSQLite(Conexion):
	''' Representa una conexión a una base de datos de SQLite '''

	def __init__ \
	(
		self,
		base_de_datos: str
	):
		# Se crea la cadena de conexión a la base de datos de SQLite y se hace la conexión
		cadena_conexion = f'sqlite:///{RUTA / base_de_datos}'
		super().__init__(cadena_conexion, SQLite.metadata)

class ConexionPostgreSQL(Conexion):
	''' Representa una conexión a una base de datos de PostgreSQL '''

	def __init__ \
	(
		self,
		host: str,
		puerto: int,
		usuario: str,
		clave: str,
		base_de_datos: str
	):
		# Se crea la cadena de conexión a la base de datos de PostgreSQL y se hace la conexión
		cadena_conexion = f'postgresql://{usuario}:{clave}@{host}:{puerto}/{base_de_datos}'
		super().__init__(cadena_conexion, PostgreSQL.metadata)

# ------------------------------ Tipos de datos -------------------------------

class TIPO_VALOR(Enum):
	'''
	Enumeración de los tipos de valores
	- CONFIRMADO
	- FALLECIDO
	'''

	CONFIRMADO = 'FECHA_SINTOMAS'
	FALLECIDO = 'FECHA_DEF'

	def __str__(self):
		return self.name

	@classmethod
	def obtener_datos(cls):
		''' Genera una lista de diccionarios para su inserción en la base de datos '''

		return [{'tipo' : str(tipo)} for tipo in cls]

class GRUPO_EDAD(Enum):
	'''
	Enumeración de los grupos de edad
	- GRUPO_18a29
	- GRUPO_30a39
	- GRUPO_40a49
	- GRUPO_50a59
	- GRUPO_60ymas
	'''

	GRUPO_18a29 = {'edad_inicial': 18, 'edad_final': 29}
	GRUPO_30a39 = {'edad_inicial': 30, 'edad_final': 39}
	GRUPO_40a49 = {'edad_inicial': 40, 'edad_final': 49}
	GRUPO_50a59 = {'edad_inicial': 50, 'edad_final': 59}
	GRUPO_60ymas = {'edad_inicial': 60, 'edad_final': None}

	def __str__(self):
		return self.name.replace('GRUPO_', '')

	@classmethod
	def obtener_datos(cls):
		''' Genera una lista de diccionarios para su inserción en la base de datos '''

		for grupo in cls:
			grupo.value['nombre'] = str(grupo)
		return [grupo.value for grupo in cls]

# ---------------------------- Modelos de PostgreSQL --------------------------

class Estado(PostgreSQL):
	__tablename__ = 'Estado'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	id_dge = sa.Column(sa.Integer, nullable = False)
	nombre = sa.Column(sa.String(50), nullable = True)

	# Relaciones
	municipios = orm.relationship('Municipio', back_populates = 'estado')

class Municipio(PostgreSQL):
	__tablename__ = 'Municipio'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	id_dge = sa.Column(sa.Integer, nullable = False)
	nombre = sa.Column(sa.String(100), nullable = True)
	fk_estado = sa.Column(sa.Integer, sa.ForeignKey('Estado.id'), nullable = False)

	# Relaciones
	estado = orm.relationship('Estado', back_populates = 'municipios')
	detalle_valores = orm.relationship('DetalleValor', back_populates = 'municipio')

class GrupoEdad(PostgreSQL):
	__tablename__ = 'GrupoEdad'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	nombre = sa.Column(sa.String(100), nullable = False)
	edad_inicial = sa.Column(sa.Integer, nullable = True)
	edad_final = sa.Column(sa.Integer, nullable = True)

	# Relaciones
	detalle_valores = orm.relationship('DetalleValor', back_populates = 'grupo_edad')

class TipoValor(PostgreSQL):
	__tablename__ = 'TipoValor'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	tipo = sa.Column(sa.String(100), nullable = False)

	# Relaciones
	detalle_valores = orm.relationship('DetalleValor', back_populates = 'tipo_valor')

class DetalleValor(PostgreSQL):
	__tablename__ = 'DetalleValor'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	fecha = sa.Column(sa.Date, nullable = False)
	fk_municipio = sa.Column(sa.Integer, sa.ForeignKey('Municipio.id'), nullable = False)
	fk_grupo_edad = sa.Column(sa.Integer, sa.ForeignKey('GrupoEdad.id'), nullable = False)
	fk_tipo_valor = sa.Column(sa.Integer, sa.ForeignKey('TipoValor.id'), nullable = False)

	# Relaciones
	municipio = orm.relationship('Municipio', back_populates = 'detalle_valores')
	grupo_edad = orm.relationship('GrupoEdad', back_populates = 'detalle_valores')
	tipo_valor = orm.relationship('TipoValor', back_populates = 'detalle_valores')
	valor_real = orm.relationship('ValorReal', back_populates = 'detalle_valor', uselist = False)
	valor_predicho = orm.relationship('ValorPredicho', back_populates = 'detalle_valor', uselist = False)

class ValorReal(PostgreSQL):
	__tablename__ = 'ValorReal'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	valor = sa.Column(sa.Integer, nullable = False)
	fk_detalle_valor = sa.Column(sa.Integer, sa.ForeignKey('DetalleValor.id'), nullable = False)

	# Relaciones
	detalle_valor = orm.relationship('DetalleValor', back_populates = 'valor_real')

class ValorPredicho(PostgreSQL):
	__tablename__ = 'ValorPredicho'

	# Atributos
	id = sa.Column(sa.Integer, primary_key = True)
	valor = sa.Column(sa.Float, nullable = False)
	fk_detalle_valor = sa.Column(sa.Integer, sa.ForeignKey('DetalleValor.id'), nullable = False)

	# Relaciones
	detalle_valor = orm.relationship('DetalleValor', back_populates = 'valor_predicho')

# ------------------------------ Modelos de SQLite ----------------------------

class RegistroLocal(SQLite):
	'''
		Modelo que representa un registro de COVID-19 en la base de datos local (SQLite)
		- Unicamente se guardan los registros con casos confirmados de COVID-19
	'''

	__tablename__ = 'RegistroLocal'

	# Atributos
	ID_REGISTRO = sa.Column(sa.String, primary_key = True)
	ENTIDAD_RES = sa.Column(sa.Integer, nullable = False)
	MUNICIPIO_RES = sa.Column(sa.Integer, nullable = False)
	FECHA_SINTOMAS = sa.Column(sa.Date, nullable = False)
	FECHA_DEF = sa.Column(sa.Date, nullable = True)
	GRUPO_EDAD = sa.Column(sa.String, nullable = True)

# ------------------------------ Modelos de Kafka -----------------------------

class RegistroKafka:
	''' Modelo que representa un registro del tópico de Kafka de casos de COVID-19 '''

	FECHA_ACTUALIZACION: Optional[str] = None
	ID_REGISTRO: Optional[str] = None
	ORIGEN: Optional[int] = None
	SECTOR: Optional[int] = None
	ENTIDAD_UM: Optional[int] = None
	SEXO: Optional[int] = None
	ENTIDAD_NAC: Optional[int] = None
	ENTIDAD_RES: Optional[int] = None
	MUNICIPIO_RES: Optional[int] = None
	TIPO_PACIENTE: Optional[int] = None
	FECHA_INGRESO: Optional[str] = None
	FECHA_SINTOMAS: Optional[str] = None
	FECHA_DEF: Optional[str] = None
	INTUBADO: Optional[int] = None
	NEUMONIA: Optional[int] = None
	EDAD: Optional[int] = None
	NACIONALIDAD: Optional[int] = None
	EMBARAZO: Optional[int] = None
	HABLA_LENGUA_INDIG: Optional[int] = None
	INDIGENA: Optional[int] = None
	DIABETES: Optional[int] = None
	EPOC: Optional[int] = None
	ASMA: Optional[int] = None
	INMUSUPR: Optional[int] = None
	HIPERTENSION: Optional[int] = None
	OTRA_COM: Optional[int] = None
	CARDIOVASCULAR: Optional[int] = None
	OBESIDAD: Optional[int] = None
	RENAL_CRONICA: Optional[int] = None
	TABAQUISMO: Optional[int] = None
	OTRO_CASO: Optional[int] = None
	TOMA_MUESTRA_LAB: Optional[int] = None
	RESULTADO_LAB: Optional[int] = None
	TOMA_MUESTRA_ANTIGENO: Optional[int] = None
	RESULTADO_ANTIGENO: Optional[int] = None
	CLASIFICACION_FINAL: Optional[int] = None
	MIGRANTE: Optional[int] = None
	PAIS_NACIONALIDAD: Optional[str] = None
	PAIS_ORIGEN: Optional[str] = None
	UCI: Optional[int] = None
	TYPE: Optional[str] = None

	def __init__(self, **kwargs):
		for campo, valor in kwargs.items():
			setattr(self, campo, valor)

	@property
	def fecha_actualizacion(self):
		return pd.Timestamp(self.FECHA_ACTUALIZACION)

	@staticmethod
	def desde_diccionario(diccionario: dict, _):
		''' Método requerido por Kafka para poder deserializar un objeto de este tipo '''

		if diccionario is None:
			return None
		return RegistroKafka(**diccionario)

# -------------------- Inicialización de la base de datos ---------------------

def cargar_tipos_valor(conexion: Conexion):
	''' Carga los tipos de valor en la base de datos '''

	conexion.sesion.execute(sa.insert(TipoValor), TIPO_VALOR.obtener_datos())

def cargar_grupos_edad(conexion: Conexion):
	''' Carga los grupos de edad en la base de datos '''

	conexion.sesion.execute(sa.insert(GrupoEdad), GRUPO_EDAD.obtener_datos())

def cargar_estados(conexion: Conexion):
	''' Carga los estados desde el diccionario de datos en la base de datos '''

	# Se cargan los estados desde el diccionario de datos de la DGE
	estados = pd.read_excel(ARCHIVO_DICCIONARIO, sheet_name = HOJA_ENTIDADES)
	estados.rename(columns = {'CLAVE_ENTIDAD': 'id_dge'}, inplace = True)
	estados['nombre'] = estados['ENTIDAD_FEDERATIVA'].str.title()

	# Se ordenan los estados por su clave de la DGE
	estados.sort_values(by = 'id_dge', inplace = True)

	# Se actualizan los estados en la base de datos
	conexion.sesion.execute(sa.insert(Estado), estados.to_dict(orient = 'records'))

def cargar_municipios(conexion: Conexion):
	''' Carga los municipios desde el diccionario de datos en la base de datos '''

	# Se cargan los municipios desde el diccionario de datos de la DGE
	municipios = pd.read_excel(ARCHIVO_DICCIONARIO, sheet_name = HOJA_MUNICIPIOS)
	municipios.rename(columns = {'CLAVE_MUNICIPIO': 'id_dge'}, inplace = True)
	municipios['nombre'] = municipios['MUNICIPIO'].str.title()

	# Se obtienen las llaves foráneas de los estados
	estados = pd.read_sql(conexion.sesion.query(Estado).statement, conexion.enlace)
	estados = estados[['id', 'id_dge']]
	estados.columns = ['fk_estado', 'CLAVE_ENTIDAD']
	municipios = pd.merge(municipios, estados, on = 'CLAVE_ENTIDAD', how = 'left')

	# Se ordenan los municipios por clave de la DGE
	municipios.sort_values(by = ['fk_estado', 'id_dge'], inplace = True)

	# Se actualizan los municipios en la base de datos
	conexion.sesion.execute(sa.insert(Municipio), municipios.to_dict(orient = 'records'))

def inicializar_base_remota(conexion: Conexion):
	''' Inicializa la base de datos remota con los datos iniciales '''

	if conexion.sesion.query(TipoValor).count() == 0:
		cargar_tipos_valor(conexion)
		cargar_grupos_edad(conexion)
		cargar_estados(conexion)
		cargar_municipios(conexion)
		conexion.sesion.commit()